import type { AxiosPromise, AxiosRequestConfig } from "axios";
import type { A, I, L, M, O, S, U } from "ts-toolbelt";
export interface Options extends AxiosRequestConfig {
    queryMethod?: 'url' | 'body';
}
export interface Stringifiable {
    toApicalypseString(): string;
}
export interface Builder<T> extends Stringifiable {
    queryFields: {
        fields?: string;
        exclude?: string;
        sort?: string;
        limit?: string;
        offset?: string;
        search?: string;
        where: string[];
    };
    queryEndpoint?: string;
    queryName?: string;
}
export interface NamedBuilder<T, N extends string> extends Omit<Builder<T>, 'queryEndpoint' | 'queryName'> {
    queryEndpoint: string;
    queryName: N;
}
export interface NarrowBuilder<T> extends Builder<T> {
    __narrow: true;
}
export interface CountBuilder<T> extends Builder<T> {
    __count: true;
}
declare type PickWith<T, K extends keyof T, X> = X extends keyof T ? Pick<T, K | X> : Pick<T, K>;
export interface Pipe<T extends object, mode extends 'result' | 'count' = 'result'> {
    <A>(...steps: (BuilderOperator<T, T> | BuilderOperatorNarrow<T, A>)[]): Stringifiable & Executor<FallbackIfUnknown<A, PickWith<T, never, 'id'>>, mode>;
}
export declare type PipeSub<T extends object, Ret extends string, mode extends 'result' | 'count' = 'result'> = {
    <A>(...steps: (BuilderOperator<T, T> | BuilderOperatorNarrow<T, A>)[]): NamedBuilder<FallbackIfUnknown<A, PickWith<T, never, 'id'>>, Ret> & (mode extends 'count' ? CountBuilder<any> : {});
};
export interface Executor<T, mode extends 'result' | 'count' = 'result'> {
    execute(url: string, options?: Options): AxiosPromise<{
        result: T[];
        count: {
            count: number;
        };
    }[mode]>;
}
export declare type ResultMultiMono<T extends NamedBuilder<any, any>> = {
    name: T["queryName"];
} & (T extends CountBuilder<any> ? {
    count: number;
} : {
    result: T extends NamedBuilder<infer S, any> ? S[] : never;
});
export interface ExecutorMulti<T extends Builder<any>[]> {
    execute(url: string, options?: Options): AxiosPromise<T extends (infer S)[] ? ResultMultiMono<S extends NamedBuilder<any, any> ? S : never>[] : never>;
}
export interface BuilderOperator<T, R> {
    (builder: Builder<T>): Builder<R>;
}
export interface BuilderOperatorNarrow<T, R> {
    (builder: Builder<T>): NarrowBuilder<R>;
}
export interface NamedBuilderOperator<T, N extends string> {
    (builder: Builder<T>): NamedBuilder<T, N>;
}
export declare type StandardOperators = '=' | '!=';
export declare type StringOperators = StandardOperators | '~';
export declare type NumbersOperatos = StandardOperators | '>=' | '>' | '<=' | '<';
export declare type AllowedValues = true | false | null;
export declare type GetOp<T> = T extends number ? NumbersOperatos : T extends string ? StringOperators : StandardOperators;
export declare enum WhereFlags {
    RAW = 1,
    NUMBER = 1,
    STRING = 2,
    STARTSWITH = 6,
    ENDSWITH = 10,
    CONTAINS = 14
}
export declare enum WhereInFlags {
    AND = 32,
    NAND = 48,
    OR = 64,
    NOR = 80,
    EXACT = 128
}
export declare type FallbackIfUnknown<T, F> = unknown extends T ? F : T;
declare type OuterKeyCast<T, K extends string> = PickOuterKey<K> & UnionKeyOf<T>;
declare type UnionKeyOf<T> = T extends Array<infer T> ? keyof T : T extends infer T ? keyof T : never;
export declare type DeepPick<T, P extends string> = T extends M.Primitive ? T : T extends Array<infer I> ? DeepPick<I, P>[] : T extends object ? string extends P ? number : PickWith<InnerPick<T, P>, OuterKeyCast<T, P>, 'id'> : never;
declare type InnerPick<T, K extends string> = {
    [key in keyof T]: DeepPick<T[key], InnerKey<Extract<key, string>, K>>;
};
export declare type FlatPath<O extends any, P extends L.List<A.Key>, X = Path<O, P>> = X extends object ? number : X;
declare type PickOuterKey<K extends string> = K extends '*' ? string : KeyHead<K>;
declare type KeyHead<K extends string> = K extends `${infer K}.${string}` ? K : K;
declare type InnerKey<key extends string, K> = [
    Extract<K, `${key}.${string}`>
] extends [`${key}.${infer K}`] ? K : never;
declare type FlatKey<O, K extends keyof O> = Exclude<O[K], null | undefined> extends Array<infer I> ? I : O[K];
declare type Flat<O> = Exclude<O, null | undefined> extends Array<infer I> ? I : O;
declare type _ExcludePrimitiveKeys<O> = O extends M.Primitive ? Omit<O, keyof O> : O;
declare type _Path<O, P extends L.List<A.Key>, It extends I.Iteration = I.IterationOf<0>> = {
    0: _Path<Flat<A.At<_ExcludePrimitiveKeys<O>, P[I.Pos<It>]>>, P, I.Next<It>>;
    1: O;
}[A.Extends<I.Pos<It>, L.Length<P>>];
export declare type Path<O extends any, P extends L.List<A.Key>> = _Path<O, P> extends infer X ? A.Cast<X, any> : never;
declare type Index = number | string;
declare type KeyToIndex<K extends A.Key, SP extends L.List<Index>> = number extends K ? L.Head<SP> : K & Index;
declare type MetaPath<O, D extends string, St extends string, SP extends L.List<Index> = [], P extends L.List<Index> = []> = {
    [K in keyof Required<O>]: Exclude<MetaPath<FlatKey<O, K>, D, St, L.Tail<SP>, [...P, KeyToIndex<K, SP>]>, string> | S.Join<[...P, KeyToIndex<K, SP>], D> | ([St] extends [never] ? never : S.Join<[...P, St], D>);
};
declare type NextPath<OP> = U.Select<O.UnionOf<Exclude<OP, string> & {}>, string>;
declare type CurrentPath<OP> = U.Select<Exclude<OP, object>, string>;
declare type ExecPath<A, SP extends L.List<Index>, D extends string, St extends string> = NextPath<Path<MetaPath<A, D, St, SP>, SP>>;
declare type HintPath<A, P extends string, SP extends L.List<Index>, Exec extends string, D extends string, St extends string> = [Exec] extends [never] ? CurrentPath<Path<MetaPath<A, D, St, SP>, SP>> extends never ? ExecPath<A, L.Pop<SP>, D, St> : CurrentPath<Path<MetaPath<A, D, St, SP>, SP>> : Exec | P;
declare type _AutoPath<A, P extends string, D extends string, St extends string, SP extends L.List<Index> = S.Split<P, D>> = HintPath<A, P, SP, ExecPath<A, SP, D, St>, D, St>;
export declare type AutoPath<O extends any, P extends string, D extends string = '.', St extends string = '*'> = _AutoPath<O, P, D, St>;
export declare type AllAutoPath<O extends object, P extends L.List<string>> = {
    [K in keyof P]: AutoPath<O, P[K] & string>;
};
export declare type NonEmptyStringList = [string, ...string[]];
export declare type ChosenPaths<P extends NonEmptyStringList> = [string] extends P ? '*' : P[number];
export {};
